//*****************************************************************************
//
//             Copyright (c) 2009 Sony Pictures Imageworks Inc.
//                            All rights reserved.
//
//  This  material  contains  the confidential and  proprietary information
//  of Sony Pictures Imageworks Inc.   and may not be disclosed,  copied or
//  duplicated  in any form,  electronic or hardcopy,  in whole  or in part,
//  without  the express prior written  consent of Sony Pictures Imageworks
//  Inc. This copyright notice does not imply publication.
//
//*****************************************************************************

#include "arnold_render.h"

#include <stdlib.h>

#include <pystring/pystring.h>
namespace pystring = SPI_pystring_v1;

void setUserMatrix(AtNode* node, Imath::M44d xform)
{
    if(!node) return;

    AtMatrix m;
    M44d_to_AtMatrix(m, xform);
    
    AiNodeDeclare(  node, "W2C", "constant MATRIX");
    AiNodeSetMatrix(node, "W2C", m);
}


void setCameraDefault(AtNode* node)
{
    setArnoldParam_Float(GroupAttribute(), "shutter_start", node);
    setArnoldParam_Float(GroupAttribute(), "shutter_end", node);
    setArnoldParam_Int(GroupAttribute(), "shutter_type", node);
    setArnoldParam_Bool(GroupAttribute(), "plane_distance", node);
    setArnoldParam_Float(GroupAttribute(), "focal_distance", node);
    setArnoldParam_Float(GroupAttribute(), "aperture_size", node);
}

namespace
{
    typedef std::map<std::string, AtNode *> CameraMap;
    CameraMap global_cameraMap;
}

AtNode *getCameraLocationByPath(const std::string &path)
{
    CameraMap::iterator iter = global_cameraMap.find(path);
    if (iter != global_cameraMap.end())
    {
        return (*iter).second;
    }
    AiMsgError("[kat] FATAL: Unable to find cached camera '%s'.", path.c_str());
    exit(-1);
}

AtNode *buildCameraLocationByPath(const std::string &path, FnScenegraphIterator rootIterator,
    int *overscan, int *displayWindowSize)
{
    CameraMap::iterator iter = global_cameraMap.find(path);
    if (iter != global_cameraMap.end())
    {
        return (*iter).second;
    }

    FnScenegraphIterator cameraIterator = rootIterator.getByPath(path);
    if (!cameraIterator.isValid())
    {
        AiMsgError("[kat] FATAL: Unable to find the camera '%s'.", path.c_str());
        exit(-1);
    }

    AtNode* cameraNode = buildCameraLocation(cameraIterator, overscan, displayWindowSize);
    if(!cameraNode)
    {
        AiMsgError("[kat] FATAL: during build of '%s'.", path.c_str());
        exit(-1);
    }

    return cameraNode;
}

AtNode* buildCameraLocation(FnScenegraphIterator cameraIterator, int *overscan, int *displayWindowSize)
{
    if (!cameraIterator.isValid()) return NULL;

    FnScenegraphIterator rootIterator = cameraIterator.getRoot();
    if (!rootIterator.isValid()) return NULL;

    std::string cameraName = cameraIterator.getFullName();

    AtNode* existingCamera = AiNodeLookUpByName(cameraName.c_str());
    if (existingCamera) return existingCamera;

    GroupAttribute cameraGeometryAttr = cameraIterator.getAttribute("geometry");
    if (!cameraGeometryAttr.isValid())
    {
        AiMsgError("[kat] FATAL: Geometry attribute no found on camera '%s'.", cameraName.c_str());
        return NULL;
    }

    // set camera type
    std::string camera_type = "persp_camera";
    StringAttribute cameraTypeAttr = cameraGeometryAttr.getChildByName("projection");
    if( cameraTypeAttr.isValid() )
    {
    	if( cameraTypeAttr.getValue() == "orthographic" )
    		camera_type = "ortho_camera";
    }

    GroupAttribute gAttr = rootIterator.getAttribute("arnoldGlobalStatements");
    AtNode* filtermapNode = NULL;
    IntAttribute planeDistanceAttr;
    FloatAttribute dofAttr;
    FloatAttribute apertureAttr;
    
    bool foundTypeAttr = false;
    
    if (gAttr.isValid())
    {
        IntAttribute typeAttr = gAttr.getChildByName("camera_type");
        if (typeAttr.isValid())
        {
            foundTypeAttr = true;
            
            int type = typeAttr.getValue();
            switch (type)
            {
                case 0:
                    camera_type = "cyl_camera";
                    break;
                case 1:
                    camera_type = "fisheye_camera";
                    break;
                case 2:
                    camera_type = "ortho_camera";
                    break;
                case 3:
                {
                    StringAttribute attr = cameraGeometryAttr.getChildByName("projection");
                    if (attr.isValid())
                    {
                        if (attr.getValue() == "orthographic")
                        {
                            camera_type = "ortho_camera";
                            break;
                        }
                    }
                }
                case 5:
                    camera_type = "bs_tdeLens";
                    break;

                default:
                    camera_type = "persp_camera";
                    break;
            }
        }
        
        //filtermap set?
        StringAttribute attrRef = gAttr.getChildByName("filtermap");
        if (attrRef.isValid())
        {
            std::string path = attrRef.getValue();
            filtermapNode = getShaderByPath(path, cameraIterator, NULL, true,"arnoldSurface");
        }
    
        planeDistanceAttr = gAttr.getChildByName("plane_distance");
        dofAttr = gAttr.getChildByName("focal_distance");
        apertureAttr = gAttr.getChildByName("aperture_size");
    }
    
    
    //fallback case is "from camera location"
    if (!foundTypeAttr)
    {
        StringAttribute attr = cameraGeometryAttr.getChildByName("projection");
        if (attr.isValid())
        {
            if (attr.getValue() == "orthographic")
            {
                camera_type = "ortho_camera";
            }
        }
    }
    
    
    //allow geometry.filtermap to override arnoldGlobalStatements.filtermap
    {
        StringAttribute attrRef = cameraGeometryAttr.getChildByName("filtermap");
        if (attrRef.isValid()) 
        {
            std::string path = attrRef.getValue();
            filtermapNode = getShaderByPath(path, cameraIterator, NULL, true,"arnoldSurface");
        }
    }

    //create cam node
    AtNode* cameraNode = AiNode(camera_type.c_str());
    if (!cameraNode)
    {
        AiMsgError("[kat] FATAL: Unable to create camera %s", cameraName.c_str());
        return NULL;
    }
    AiNodeSetStr(cameraNode, "name", cameraName.c_str());

    //hookup filtermap if we have one
    if (filtermapNode)
    {
        AiNodeSetPtr(cameraNode, "filtermap", filtermapNode);
    }
    
    //clipping plane correction?
    if (planeDistanceAttr.isValid())
    {
        AiNodeSetBool(cameraNode, "plane_distance", planeDistanceAttr.getValue());
    }
   
    //dof
    if (dofAttr.isValid() && apertureAttr.isValid())
    {
        //TODO: MB
        float dof = dofAttr.getValue();
        float aw = apertureAttr.getValue();
        if (dof != 0.0f && aw != 0.0f) 
        {
            AiNodeSetArray(cameraNode, "focal_distance", AiArrayConvert(1, 1, AI_TYPE_FLOAT, &dof));
            AiNodeSetArray(cameraNode, "aperture_size", AiArrayConvert(1, 1, AI_TYPE_FLOAT, &aw));
            
            IntAttribute intAttr;
            FloatAttribute floatAttr;

            intAttr = gAttr.getChildByName("aperture_nsides");
            if (intAttr.isValid())
            {
                AiNodeSetInt(cameraNode, "aperture_nsides", intAttr.getValue());
            }
            floatAttr = gAttr.getChildByName("aperture_ratio");
            if (floatAttr.isValid())
            {
                AiNodeSetFlt(cameraNode, "aperture_ratio", floatAttr.getValue());
            }
            floatAttr = gAttr.getChildByName("aperture_relax");
            if (floatAttr.isValid())
            {
                AiNodeSetFlt(cameraNode, "aperture_relax", floatAttr.getValue());
            }
            floatAttr = gAttr.getChildByName("aperture_rotation");
            if (floatAttr.isValid())
            {
                AiNodeSetFlt(cameraNode, "aperture_rotation", floatAttr.getValue());
            }
        }
    }
    
    //screen window 
    {
        DoubleAttribute leftAttr = cameraGeometryAttr.getChildByName("left");
        DoubleAttribute rightAttr = cameraGeometryAttr.getChildByName("right");
        DoubleAttribute bottomAttr = cameraGeometryAttr.getChildByName("bottom");
        DoubleAttribute topAttr = cameraGeometryAttr.getChildByName("top");
        if (!leftAttr.isValid() || !rightAttr.isValid() ||
            !bottomAttr.isValid() || !topAttr.isValid())
        {
            AiMsgError("[kat] FATAL: Missing a required attribute in the geometry group on camera '%s'.",
                cameraName.c_str());
            return NULL;
        }

        double left = leftAttr.getValue();
        double right = rightAttr.getValue();
        double bottom = bottomAttr.getValue();
        double top = topAttr.getValue();

        // adjust screen window for overscan
        if (overscan != 0 && displayWindowSize != 0)
        {
            if (displayWindowSize[0] == 0 || displayWindowSize[1] == 0)
            {
                AiMsgWarning("[kat] Ignoring overscan in screen window calcuation, since display window size is 0.");
            }

            int displayWindow[4] = {0, 0, displayWindowSize[0], displayWindowSize[1]};

            double displayWindow_w2 = (displayWindow[2] - displayWindow[0]) / 2.0;
            double displayWindow_h2 = (displayWindow[3] - displayWindow[1]) / 2.0;
            double overscanMult[4] = {(displayWindow_w2+overscan[0])/displayWindow_w2,
                                      (displayWindow_h2+overscan[1])/displayWindow_h2,
                                      (displayWindow_w2+overscan[2])/displayWindow_w2,
                                      (displayWindow_h2+overscan[3])/displayWindow_h2};
            double screen_window_cx = (left + right) / 2.0;
            double screen_window_cy = (bottom + top) / 2.0;

            left = screen_window_cx + (left-screen_window_cx)*overscanMult[0];
            bottom = screen_window_cy + (bottom-screen_window_cy)*overscanMult[1];
            right = screen_window_cx + (right-screen_window_cx)*overscanMult[2];
            top = screen_window_cy + (top-screen_window_cy)*overscanMult[3];
        }

        if (camera_type == "ortho_camera")
        {
            DoubleAttribute orthoWidthAttr =
                cameraGeometryAttr.getChildByName("orthographicWidth");
            
            if (orthoWidthAttr.isValid() && orthoWidthAttr.getValue() > 0)
            {
                double orthoWidth = orthoWidthAttr.getValue();
                
                double w = (right-left);
                double h = (top-bottom);
                
                left *= 1.0/w * orthoWidth;
                right *= 1.0/w * orthoWidth;
                
                top *= 1.0/h * orthoWidth;
                bottom *= 1.0/h * orthoWidth;
            }
        }
        
        AiNodeSetPnt2(cameraNode, "screen_window_min", left, bottom);
        AiNodeSetPnt2(cameraNode, "screen_window_max", right, top);
    }

    //near
    {
        DoubleAttribute nearAttr = cameraGeometryAttr.getChildByName("near");
        if (!nearAttr.isValid())
        {
            AiMsgError("[kat] FATAL: 'geometry.near' not found on camera '%s'.", cameraName.c_str());
            return NULL;
        }
        AiNodeSetFlt(cameraNode, "near_clip", nearAttr.getValue());
    }

    //far
    {
        DoubleAttribute farAttr = cameraGeometryAttr.getChildByName("far");
        if (!farAttr.isValid())
        {
            AiMsgError("[kat] FATAL: 'geometry.far' not found on camera '%s'.", cameraName.c_str());
            return NULL;
        }
        AiNodeSetFlt(cameraNode, "far_clip", farAttr.getValue());
    }
    
    //shutter
    {
        
        /*FloatAttribute shutterStartAttr = rootIterator.getAttribute("arnoldGlobalStatements.shutter_start");
        if (shutterStartAttr.isValid())
        {
            AiNodeSetFlt(cameraNode, "shutter_start", shutterStartAttr.getValue());
            SetGlobalShutterOpen(shutterStartAttr.getValue());
        }
        
        FloatAttribute shutterEndAttr = rootIterator.getAttribute("arnoldGlobalStatements.shutter_end");
        if (shutterEndAttr.isValid())
        {
            AiNodeSetFlt(cameraNode, "shutter_end", shutterEndAttr.getValue());
            SetGlobalShutterClose(shutterEndAttr.getValue());
        }*/
        
        //TRANSITIONAL: override with possible global shutter value
        FloatAttribute shutterOpenAttr = rootIterator.getAttribute("renderSettings.shutterOpen");
        if (shutterOpenAttr.isValid())
        {
            AiNodeSetFlt(cameraNode, "shutter_start", shutterOpenAttr.getValue());
            SetGlobalShutterOpen(shutterOpenAttr.getValue());
        }
        FloatAttribute shutterCloseAttr = rootIterator.getAttribute("renderSettings.shutterClose");
        if (shutterCloseAttr.isValid())
        {
            AiNodeSetFlt(cameraNode, "shutter_end", shutterCloseAttr.getValue());
            SetGlobalShutterClose(shutterCloseAttr.getValue());
        }
        
        IntAttribute shutterTypeAttr = rootIterator.getAttribute("arnoldGlobalStatements.shutter_type");
        if (shutterTypeAttr.isValid())
        {
            AiNodeSetInt(cameraNode, "shutter_type", shutterTypeAttr.getValue());
        }
    }
    
    {
        std::map<float, Imath::M44d> xformSamples;
        applyTransform(cameraNode, cameraIterator, &xformSamples);
        
        // give mark his W2C
        setUserMatrix(cameraNode, xformSamples[0]);
        
        //fov (output at same (interpolated) time samples as xform)
        {
            DoubleAttribute fovAttr = cameraGeometryAttr.getChildByName("fov");
            if (!fovAttr.isValid())
            {
                AiMsgError("[kat] FATAL: Geometry attribute missing the fov attribute on camera '%s'.",
                    cameraName.c_str());
                return NULL;
            }
            
            std::map<float, Imath::M44d>::const_iterator xformSampleIter;
            std::vector<float> fovValues;
            
            for (xformSampleIter=xformSamples.begin(); xformSampleIter != xformSamples.end(); ++xformSampleIter)
            {
                DoubleConstVector fovSample = fovAttr.getInterpSample((*xformSampleIter).first);
                fovValues.push_back(fovSample[0]);
            }

            AiNodeSetArray(cameraNode, "fov", AiArrayConvert(1, fovValues.size(), AI_TYPE_FLOAT, &fovValues[0]));
        }
    }

    global_cameraMap[cameraName] = cameraNode;

    AiMsgInfo("[kat] made camera %s", cameraName.c_str());

    return cameraNode;
}

std::string getDefaultRenderCameraName(FnScenegraphIterator rootIterator)
{
    StringAttribute cameraNameAttr = rootIterator.getAttribute("renderSettings.cameraName");
    return cameraNameAttr.getValue("/root/world/cam/camera", false);
}

void buildRenderCamera(FnScenegraphIterator rootIterator, int *overscan, int *displayWindowSize)
{
    std::string cameraName = getDefaultRenderCameraName(rootIterator);

    AtNode* cameraNode = buildCameraLocationByPath(cameraName, rootIterator, overscan, displayWindowSize);
    if(!cameraNode)
    {
        AiMsgError("[kat] FATAL: during build of '%s'.", cameraName.c_str());
        exit(-1);
    }

    // Point the option block at this camera.
    AiNodeSetPtr(AiUniverseGetOptions(), "camera", cameraNode);
    AiMsgInfo("[kat] render camera created.");
}

void buildCameras(FnScenegraphIterator worldIterator)
{
    StringAttribute cameraListAttr = worldIterator.getAttribute("globals.cameraList");
    if (!cameraListAttr.isValid())
    {
        AiMsgInfo("[kat] No globals.cameraList found at /root/world.  Only the render camera will be preloaded.");
        return;
    }

    StringConstVector cameraPaths = cameraListAttr.getNearestSample(0.0f);
    for (StringConstVector::const_iterator iter = cameraPaths.begin();
         iter != cameraPaths.end(); ++iter)
    {
        std::string cameraPath = (*iter);
        if (pystring::startswith(cameraPath, "/root"))
        {
            cameraPath = cameraPath.substr(5);
        }
        FnScenegraphIterator cameraIterator = worldIterator.getByPath(cameraPath);
        if (!cameraIterator.isValid())
        {
            AiMsgInfo("[kat] The camera '%s' could not be found in the scene.", (*iter).c_str());
            continue;
        }

        buildCameraLocation(cameraIterator);
    }
}
